// ======================================================================
//
// Vector.h
// jeff grills
//
/// Vector/Overview
///
/// The Vector class provides a standard 3 dimensional mathematical vector.
///
/// This class contains some public Vector constants within its names space that are very useful:
/// unitX, unitY, unitZ, unitNegativeX, unitNegativeY, unitNegatizeZ, and zero.
//
// copyright 1998 origin systems inc
//
// ======================================================================

#ifndef VECTOR_H
#define VECTOR_H

// ======================================================================

class Vector
{
public:

	static const Vector unitX;
	static const Vector unitY;
	static const Vector unitZ;
	static const Vector unitNegativeX;
	static const Vector unitNegativeY;
	static const Vector unitNegativeZ;
	static const Vector zero;
	static const real   normalizeThreshold;

public:

	enum
	{
		SAVE_SIZE = 3
	};

	struct Save
	{
		real save[SAVE_SIZE];
	};

public:

	real x;
	real y;
	real z;

public:

	Vector(void);
	Vector(real newX, real newY, real newZ);

	// assignment
	void    set(real newX, real newY, real newZ);
	void    makeZero(void);
	
	// comparison operators
	bool    operator ==(const Vector &v2) const;
	bool    operator !=(const Vector &v2) const;
	bool    isZero(void) const;

	// misc functions
	bool    normalize(void);

	// random vector functions
	Vector &randomUnit(void);
	Vector &random(real radius);

	// vector math routines
	void    add(const Vector &u, const Vector &v);
	void    add(real x, real y, real z);
	void    sub(const Vector &u, const Vector &v);
	void    sub(real x, real y, real z);
	void    mul(const Vector &u, real i);
	void    div(const Vector &u, real i);

	// operator vector math routines
	Vector &operator +=(const Vector &v2);
	Vector &operator -=(const Vector &v2);
	Vector &operator *=(real i);
	Vector &operator /=(real i);

	// cross product tougines
	void    cross(const Vector &u, const Vector &v);
	void    cross(real ux, real uy, real uz, real vx, real vy, real vz);

	// dot product routines
	real    dot(const Vector &v) const;
	real    dot(real vx, real vy, real vz) const;
	void    dot(const Vector *v, real *dotProduct, int count) const;

	// magnitude routines
	real    magnitudeSquared(void) const;
	real    magnitude(void) const;
	real    approximateMagnitude(void) const;
	real    magnitudeSquaredBetween(const Vector & v) const;
	real    magnitudeBetween(const Vector &v ) const;
	real    approximateMagnitudeBetween(const Vector &v) const;

	// vector orientation angles
	real    theta(void) const;
	real    phi(void) const;
	real    phiMajor(void) const;

	// vector building
	void    makePerpendicular(const Vector &to);
	void    reflectOutgoing(const Vector &incident, const Vector &normal);
	void    reflectIncoming(const Vector &incident, const Vector &normal);
	void    midpoint(const Vector &u, const Vector &v);
	void    reverse(void);
	void    reverse(const Vector &u);

	void   writeSave(Save *save);
	void   readSave(const Save &save);
};

// ======================================================================
/// Vector/Functions/Construction/Vector(void)
///
/// Construct a vector
///
/// Remarks:
///
/// The vector is initialized with its x, y, and z components all set to zero.

inline Vector::Vector(void)
: x(CONST_REAL(0)), y(CONST_REAL(0)), z(CONST_REAL(0))
{	
}

// ------------------------------------------------------------------------
/// Vector/Functions/Construction/Vector(real x, real y, real z)
///
/// Construct a vector
///
/// Parameters:
///
/// x The new value of the X component of the 3d vector
///
/// y The new value of the Y component of the 3d vector
///
/// z The new value of the Z component of the 3d vector
/// 
/// Remarks:
///
/// This vector is initialized with its x, y, and z components set to the arguments.

inline Vector::Vector(real newX, real newY, real newZ)
: x(newX), y(newY), z(newZ)
{
}

// ------------------------------------------------------------------------
/// Vector/Functions/Assignment/void set(real x, real y, real z)
///
/// Changes the value of the vector
///
/// Parameters:
///
/// x The new value of the X component of the 3d vector
///
/// y The new value of the Y component of the 3d vector
///
/// z The new value of the Z component of the 3d vector
/// 
/// Remarks:
///
/// This vector modified so that its x, y, and z components are set to the arguments.

inline void Vector::set(real newX, real newY, real newZ)
{
	x = newX;
	y = newY;
	z = newZ;
}

// ----------------------------------------------------------------------
/// Vector/Functions/Magnitude/real magnitudeSquared(void) const
///
/// Return Value:
///
/// The square of the magnitude of this vector
/// 
/// Remarks:
///
/// This routine returns the square of the magnitude of the vector.  It is significantly more
/// efficient than magnitude() because it does not need to calculate a square root.
///
/// In many applications, the square of the magnitude can be used instead of the actual magnitude.
/// If the magnitudes of two vectors need compared, and the magnitudes  are greater that 1.0, you
/// can simply compare the squares of the magnitudes.

inline real Vector::magnitudeSquared(void) const
{
	return (sqr(x) + sqr(y) + sqr(z));
}

// ----------------------------------------------------------------------
/// Vector/Functions/Magnitude/real magnitude(void) const
///
/// Return Value:
///
/// The magnitude of this vector
/// 
/// Remarks:
///
/// This routine is slow because it must calculate a square root.
///
/// See also approximateMagnitude() and magnitudeSquared()

inline real Vector::magnitude(void) const
{
	return sqrt(magnitudeSquared());
}

// ------------------------------------------------------------------------
/// Vector/Functions/Magnitude/real magnitudeSquaredBetween(const Vector &v) const
///
/// Returns the square of the magnitude of the vector between the point at this vector and the point at vector v
///
/// Return value:
///
/// The square of the magnitude of the vector between the point at this vector and the point at vector v
///
/// Parameters:
///
/// v The point to determine the square of the magnitude to
///
/// Remarks:
///
/// This routine computes the square of the magnitude from this vector to vector v.  It effectively (and
/// efficiently) computes the delta vector, and then computes the square of the magnitude of that.

inline real Vector::magnitudeSquaredBetween(const Vector &v) const
{
	return (sqr(x - v.x) + sqr(y - v.y) + sqr(z - v.z));
}

// ------------------------------------------------------------------------
/// Vector/Functions/Magnitude/real magnitudeBetween(const Vector &v) const
///
/// Returns the magnitude of the vector between the point at this vector and the point at vector v
///
/// Return value:
///
/// The magnitude of the vector between the point at this vector and the point at vector v
///
/// Parameters:
///
/// v The point to determine the magnitude to
///
/// Remarks:
///
/// This routine is slow because it must calculate a square root.  In many applications, the
/// square of the magnitude can be used instead.
///
/// This routine computes the magnitude from this vector to vector v.  It effectively (and efficiently)
/// computes the delta vector, and then computes the square of the magnitude of that.
///
/// See approximateMagnitudeBetween(), magnitudeSquaredBetween().

inline real Vector::magnitudeBetween(const Vector &v) const
{
	return sqrt(magnitudeSquaredBetween(v));
}

// ------------------------------------------------------------------------
/// Vector/Functions/Building/void reverse(void)
///
/// Reverses the direction of this vector
///
/// Remarks:
///
/// This routine negates the x, y, and z components of this vector, effectively reversing its direction.

inline void Vector::reverse(void)
{
	x = -x;
	y = -y;
	z = -z;
}

// ------------------------------------------------------------------------
/// Vector/Functions/Building/void reverse(const Vector &v)
///
/// Reverses the direction of vector V and stores the result in this vector
///
/// Parameters:
///
/// v the vector to reverse
///
/// Remarks:
///
/// This routine negates the x, y, and z components of vector v, storing the resulting values in this
/// vector, effectively reversing its direction.

inline void Vector::reverse(const Vector &v)
{
	x = -v.x;
	y = -v.y;
	z = -v.z;
}

// ------------------------------------------------------------------------
/// Vector/Functions/Assignment/void makeZero(void)
///
/// Sets the x, y, and z components of this vector to zero
///
/// Remarks:
///
/// This is the prefered way to set a vector to the zero vector.

inline void Vector::makeZero(void)
{
	x = CONST_REAL(0);
	y = CONST_REAL(0);
	z = CONST_REAL(0);
}

// ------------------------------------------------------------------------
/// Vector/Functions/Comparison/bool isZero(void) const
///
/// Tests this vector to see if it is the zero vector
///
/// Return value:
///
/// True if all components were zero, false otherwise
///
/// Remarks:
///
/// Since the vector class uses real numbers for its components, it is unlikely that a series of mathematical
/// operations will result in exactly the zero vector.

inline bool Vector::isZero(void) const
{
	return (x == CONST_REAL(0) && y == CONST_REAL(0) && z == CONST_REAL(0));
}

// ------------------------------------------------------------------------
/// Vector/Functions/Math/Vector &operator -=(const Vector &v)
///
/// This method subtracts vector v from this vector
///
/// Return value:
///
/// A reference to this vector
///
/// Parameters:
///
/// v The vector to subtract from this vector
///
/// Remarks:
///
/// Vector v is subtracted from this vector in a component-wise manner.
///
/// This overloaded operator does not create any temporary variables.

inline Vector &Vector::operator -=(const Vector &v2)
{
	x -= v2.x;
	y -= v2.y;
	z -= v2.z;
	return *this;
}

// ------------------------------------------------------------------------
/// Vector/Functions/Math/Vector &operator +=(const Vector &v)
///
/// This method adds vector v from this vector
///
/// Return value:
///
/// A reference to this vector
///
/// Parameters:
///
/// v The vector to add from this vector
///
/// Remarks:
///
/// This vector and vector v are added in a component-wise manner.
///
/// This overloaded operator does not create any temporary variables.

inline Vector &Vector::operator +=(const Vector &v2)
{
	x += v2.x;
	y += v2.y;
	z += v2.z;
	return *this;
}

// ------------------------------------------------------------------------
/// Vector/Functions/Math/Vector &operator *=(real r)
///
/// This method multiplies this vector by the specified scalar
///
/// Return value:
///
/// A reference to this vector
///
/// Parameters:
///
/// r The scalar to multiply by
///
/// Remarks:
///
/// The argument r is multiplied against each component.
///
/// This overloaded operator does not create any temporary variables.

inline Vector &Vector::operator *=(real f)
{
	x *= f;
	y *= f;
	z *= f;
	return *this;
}

// ------------------------------------------------------------------------
/// Vector/Functions/Math/Vector &operator /=(real r)
///
/// This method divides this vector by a scalar
///
/// Return value:
///
/// A reference to this vector
///
/// Parameters:
///
/// r The scalar to divide by
///
/// Remarks:
///
/// The argument r is reciprocated once, and then multiplied against each component.
///
/// This overloaded operator does not create any temporary vector variables.

inline Vector &Vector::operator /=(real f)
{
	return (*this *= RECIP(f));
}

// ------------------------------------------------------------------------
/// Vector/Functions/Comparison/bool operator ==(const Vector &v) const
///
/// Determines if two vectors have identical values
///
/// Return value:
///
/// Returns true if the two vectors have the same value, false otherwise.
///
/// Parameters:
///
/// v The vector to compare against
///
/// Remarks:
///
/// Since the vector class uses real numbers for its components, it is unlikely that a series of mathematical
/// operations will result in exactly the zero vector.

inline bool Vector::operator ==(const Vector &v2)	const
{
	return (x == v2.x && y == v2.y && z == v2.z); //lint !e777 // testing float equality
}

// ------------------------------------------------------------------------
/// Vector/Functions/Comparison/bool operator !=(const Vector &v) const
///
/// Determines if two vectors have different values
///
/// Return value:
///
/// Returns true if the two vectors have different values, false otherwise.
///
/// Parameters:
///
/// v The vector to compare against
///
/// Remarks:
///
/// Since the vector class uses real numbers for its components, it is unlikely that a series of mathematical
/// operations will result in exactly the zero vector.

inline bool Vector::operator !=(const Vector &v2)	const
{
	return (x != v2.x || y != v2.y || z != v2.z);  //lint !e777 // testing float equality
}

// ----------------------------------------------------------------------
/// Vector/Functions/Math/void add(const Vector &u, const Vector &v)
///
/// Adds the two arguments together and stores the result in this vector
///
/// Parameters:
///
/// u The first vector to add
///
/// v The second vector to add
///
/// Remarks:
///
/// The overloaded operator version of add is not provided because it may create a temporary vector.
///
/// The expression "a = b + c" is equivalent to "a.add(b, c)"

inline void Vector::add(const Vector &u, const Vector &v)
{
	x = u.x + v.x;
	y = u.y + v.y;
	z = u.z + v.z;
}

// ----------------------------------------------------------------------
/// Vector/Functions/Math/void add(real x, real y, real z)
///
/// Adds the three components to this vector
///
/// Parameters:
///
/// x The x component to add
///
/// y The y component to add
///
/// z The z component to add
///
/// Remarks:
///
/// This version of the routine is provided to reduce the number of temporary vectors
/// that need to be created.

inline void Vector::add(real addX, real addY, real addZ)
{
	x += addX;
	y += addY;
	z += addZ;
}

// ----------------------------------------------------------------------
/// Vector/Functions/Math/void sub(const Vector &u, const Vector &v)
///
/// Adds the two arguments together and stores the result in this vector
///
/// Parameters:
///
/// u the vector to subtract from
///
/// v the vector to subract
///
/// Remarks:
///
/// The overloaded operator version of add is not provided because it may create a temporary vector.
///
/// The expression "a = b - c" is equivalent to "a.sub(b, c)"

inline void Vector::sub(const Vector &u, const Vector &v)
{
	x = u.x - v.x;
	y = u.y - v.y;
	z = u.z - v.z;
}

// ----------------------------------------------------------------------
/// Vector/Functions/Math/void sub(real x, real y, real z)
///
/// Subtracts the three components from this vector
///
/// Parameters:
///
/// x The x component to subtract
///
/// y The y component to subtract
///
/// z The z component to subtract
///
/// Remarks:
///
/// This version of the routine is provided to reduce the number of temporary vectors
/// that need to be created.

inline void Vector::sub(real subX, real subY, real subZ)
{
	x -= subX;
	y -= subY;
	z -= subZ;
}

// ----------------------------------------------------------------------
/// Vector/Functions/Building/void midpoint(const Vector &u, const Vector &v)
///
/// Sets this vector to the midpoint of the two argument points
///
/// Parameters:
///
/// u The first end point
///
/// v The second end point
///
/// Remarks:
///
/// This routine averages the x, y, and z components of the two specified vectors.
///
/// It uses multiplication by one-half instead of division for efficiency.

inline void Vector::midpoint(const Vector &u, const Vector &v)
{
	x = (u.x + v.x) * CONST_REAL(0.5);
	y = (u.y + v.y) * CONST_REAL(0.5);
	z = (u.z + v.z) * CONST_REAL(0.5);
}

// ----------------------------------------------------------------------
/// Vector/Functions/Math/void mul(const Vector &u, real f)
///
/// Sets this vector to the value of u scaled by f
///
/// Parameters:
///
/// u The vector to scale
///
/// f The amount to scale by
///
/// Remarks:
///
/// This routine multiplies all three components by the specified scalar value.

inline void Vector::mul(const Vector &u, real f)
{
	x = u.x * f;
	y = u.y * f;
	z = u.z * f;
}

// ----------------------------------------------------------------------
/// Vector/Functions/Math/void div(const Vector &u, real f)
///
/// Sets this vector to the value of u scaled by 1/f
///
/// Parameters:
///
/// u The vector to scale
///
/// f The inverse of the amount to scale by
///
/// Remarks:
///
/// This routine reciprocated the value f once, and then multiplies each component by it for
/// efficiency.

inline void Vector::div(const Vector &u, real f)
{
	mul(u, RECIP(f));
}

// ----------------------------------------------------------------------
/// Vector/Functions/Math/void cross(real ux, real uy, real uz, real vx, real vy, real vz)
///
/// Sets this vector value of u cross v
///
/// Parameters:
///
/// ux The x component of the first vector
///
/// ux The y component of the first vector
///
/// ux The z component of the first vector
///
/// vx The x component of the second vector
///
/// vx The y component of the second vector
///
/// vx The z component of the second vector
///
/// Remarks:
///
/// This version of this routine is provided to reduce the number of temporary vectors
/// that need to be created.

inline void Vector::cross(real ux, real uy, real uz, real vx, real vy, real vz)
{
	x = uy*vz - uz*vy;
	y = uz*vx - ux*vz;
	z = ux*vy - uy*vx;
}

// ----------------------------------------------------------------------
/// Vector/Functions/Math/void cross(const Vector &u, const Vector &v)
///
/// Sets this vector value of u cross v
///
/// Parameters:
///
/// u The first vector
///
/// v The vector to cross

inline void Vector::cross(const Vector &u, const Vector &v)
{
	cross(u.x, u.y, u.z, v.x, v.y, v.z);
}

// ------------------------------------------------------------------------
/// Vector/Functions/Math/real dot(real vx, real vy, real vz)
///
/// Computes the dot product between this vector and the component-wise specified vector  
///
/// Return Value:
///
/// The dot product of the two vectors
///
/// Parameters:
///
/// vx The x component of the vector to dot against
///
/// vy The y component of the vector to dot against
///
/// vz The z component of the vector to dot against
///
/// Remarks:
///
/// This version of this routine is provided to reduce the number of temporary vectors
/// that need to be created.

inline real Vector::dot(real vx, real vy, real vz) const
{
	return (x * vx) + (y * vy) + (z * vz);
}

// ------------------------------------------------------------------------
/// Vector/Functions/Math/real dot(const vector &v)
///
/// Computes the dot product between this vector and the specified vector  
///
/// Return Value:
///
/// The dot product of the two vectors
///
/// Parameters:
///
/// v The vector to compute the doc product

inline real Vector::dot(const Vector &v)	const
{
	return dot(v.x, v.y, v.z);
}

// ----------------------------------------------------------------------
/// Vector/Functions/Building/void reflectOutgoing(const vector &incident, const vector &normal)
///
/// Computes the reflection of the incident vector around the normal
///
/// Parameters:
///
/// incident A vector incident to the to the surface, pointing towards the origin
///
/// normal The normal around which to reflect
///
/// Remarks:
///
/// This routine will compute the reflection of the incident vector around the normal.
/// It assumes that the intersection point is at the origin.
///
/// It would be really nice to have a way to include a bitmap image, but since the help
/// program does not have that functionality, a simple ascii picture will have to do.  'i'
/// is the incident vector, the normal vector is 'N', and 'r' is the resultant vector.
/// The arrow heads indicate the direction the normal is facing.
///
/// Example:
///
///      N
/// i <  ^  > r
///     \|/
/// -----0------

inline void Vector::reflectOutgoing(const Vector &incident, const Vector &normal)
{
	*this  = normal;
	*this *= (CONST_REAL(2) * dot(incident));
	*this -= incident;
}

// ----------------------------------------------------------------------
/// Vector/Functions/Building/void reflectIncoming(const vector &incident, const vector &normal)
///
/// Computes the reflection of the incident vector around the normal
///
/// Parameters:
///
/// incident A vector incident to the to the surface, pointing towards the origin
///
/// normal The normal around which to reflect
///
/// Remarks:
///
/// This routine will compute the reflection of the incident vector around the normal.
/// It assumes that the intersection point is at the origin.
///
/// It would be really nice to have a way to include a bitmap image, but since the help
/// program does not have that functionality, a simple ascii picture will have to do.  'i'
/// is the incident vector, the normal vector is 'N', and 'r' is the resultant vector.
/// The arrow heads indicate the direction the normal is facing.
///
/// Example:
///
///      N
/// i \  ^  > r
///     >|/
/// -----+------

inline void Vector::reflectIncoming(const Vector &incident, const Vector &normal)
{
	*this  = normal;
	*this *= (CONST_REAL(-2) * dot(incident));
	*this += incident;
}

// ----------------------------------------------------------------------

inline void Vector::writeSave(Save *save)
{
	save->save[0] = x;
	save->save[1] = y;
	save->save[2] = z;
}

// ----------------------------------------------------------------------

inline void Vector::readSave(const Save &save)
{
	x = save.save[0];
	y = save.save[1];
	z = save.save[2];
}

// ======================================================================

#endif
